@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2012-2019 Meltytech, LLC
+ * Copyright (c) 2012-2021 Meltytech, LLC
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,6 +19,8 @@
 #include "ui_playlistdock.h"
 #include "dialogs/durationdialog.h"
 #include "dialogs/filedatedialog.h"
+#include "dialogs/longuitask.h"
+#include "dialogs/slideshowgeneratordialog.h"
 #include "mainwindow.h"
 #include "settings.h"
 #include "shotcut_mlt_properties.h"
@@ -27,6 +29,8 @@
 #include "widgets/playlistlistview.h"
 #include "util.h"
 #include "commands/playlistcommands.h"
+#include "proxymanager.h"
+#include "qmltypes/qmlapplication.h"
 #include <Logger.h>
 
 #include <QMenu>
@@ -36,6 +40,8 @@
 #include <QHeaderView>
 #include <QKeyEvent>
 #include <QDir>
+#include <QGuiApplication>
+#include <QClipboard>
 
 static const int kInOutChangedTimeoutMs = 100;
 
@@ -43,7 +49,7 @@ class TiledItemDelegate : public QStyledItemDelegate
 {
     Q_OBJECT
 public:
-    TiledItemDelegate(QAbstractItemView * view, QWidget *parent = 0)
+    TiledItemDelegate(QAbstractItemView * view, QWidget *parent = nullptr)
          : QStyledItemDelegate(parent),
          m_view(view)
     {
@@ -57,7 +63,7 @@ public:
         const QImage thumb = index.data(Qt::DecorationRole).value<QImage>();
         const QString setting = Settings.playlistThumbnails();
         const int lineHeight = painter->fontMetrics().height();
-        const bool roomEnoughForAllDetails = lineHeight * 4 < thumb.height();
+        const bool roomEnoughForAllDetails = lineHeight * 5 < thumb.height();
         const QFont oldFont = painter->font();
         QFont boldFont(oldFont);
         boldFont.setBold(true);
@@ -84,7 +90,7 @@ public:
         painter->setFont(oldFont);
 
         QRect centeredTextRect = option.rect;
-        centeredTextRect.setHeight(lineHeight * (roomEnoughForAllDetails ? 4 : 2));
+        centeredTextRect.setHeight(lineHeight * (roomEnoughForAllDetails ? 5 : 3));
         centeredTextRect.moveCenter(option.rect.center());
 
         QRect textRect = centeredTextRect;
@@ -93,9 +99,14 @@ public:
         QPoint textPoint = textRect.topLeft();
         textPoint.setY(textPoint.y() + lineHeight);
         painter->setFont(boldFont);
+        QStringList nameParts = index.data(Qt::DisplayRole).toString().split('\n');
         painter->drawText(textPoint,
-                painter->fontMetrics().elidedText(index.data(Qt::DisplayRole).toString(), Qt::ElideMiddle, textRect.width()));
+                painter->fontMetrics().elidedText(nameParts.first(), Qt::ElideMiddle, textRect.width()));
         painter->setFont(oldFont);
+        if (nameParts.size() > 1) {
+            textPoint.setY(textPoint.y() + lineHeight);
+            painter->drawText(textPoint, nameParts.last());
+        }
 
         textPoint.setY(textPoint.y() + lineHeight);
         painter->drawText(textPoint, tr("Duration: %1").arg(index.data(PlaylistModel::FIELD_DURATION).toString()));
@@ -137,6 +148,9 @@ PlaylistDock::PlaylistDock(QWidget *parent) :
     LOG_DEBUG() << "begin";
     ui->setupUi(this);
     toggleViewAction()->setIcon(windowIcon());
+    ui->actionPlayAfterOpen->setChecked(Settings.playlistAutoplay());
+
+    ui->stackedWidget->setCurrentIndex(0);
 
     m_iconsView = new PlaylistIconView(this);
     ui->listView->parentWidget()->layout()->addWidget(m_iconsView);
@@ -203,7 +217,9 @@ PlaylistDock::PlaylistDock(QWidget *parent) :
 
     updateViewModeFromActions();
 
+    m_inChangedTimer.setInterval(kInOutChangedTimeoutMs);
     m_inChangedTimer.setSingleShot(true);
+    m_outChangedTimer.setInterval(kInOutChangedTimeoutMs);
     m_outChangedTimer.setSingleShot(true);
     connect(&m_inChangedTimer, SIGNAL(timeout()), this, SLOT(onInTimerFired()));
     connect(&m_outChangedTimer, SIGNAL(timeout()), this, SLOT(onOutTimerFired()));
@@ -228,6 +244,30 @@ int PlaylistDock::position()
     return result;
 }
 
+void PlaylistDock::replaceClipsWithHash(const QString& hash, Mlt::Producer& producer)
+{
+    QList<Mlt::Producer> producers;
+    for (int i = 0; i < m_model.rowCount(); ++i) {
+        QScopedPointer<Mlt::Producer> clip(m_model.playlist()->get_clip(i));
+        if (Util::getHash(clip->parent()) == hash) {
+            clip->set(kPlaylistIndexProperty, i + 1);
+            producers << *clip;
+        }
+    }
+    auto n = producers.size();
+    if (n > 1) {
+        MAIN.undoStack()->beginMacro(tr("Replace %n playlist items", nullptr, n));
+    }
+    for (auto& clip : producers) {
+        Util::applyCustomProperties(producer, clip.parent(), clip.get_in(), clip.get_out());
+        MAIN.undoStack()->push(
+            new Playlist::ReplaceCommand(m_model, MLT.XML(&producer), clip.get_int(kPlaylistIndexProperty) - 1));
+    }
+    if (n > 1) {
+        MAIN.undoStack()->endMacro();
+    }
+}
+
 void PlaylistDock::incrementIndex()
 {
     QModelIndex index = m_view->currentIndex();
@@ -260,15 +300,19 @@ void PlaylistDock::setIndex(int row)
 void PlaylistDock::moveClipUp()
 {
     int row = m_view->currentIndex().row();
-    if (row > 0)
+    if (row > 0) {
         MAIN.undoStack()->push(new Playlist::MoveCommand(m_model, row, row - 1));
+        resetPlaylistIndex();
+    }
 }
 
 void PlaylistDock::moveClipDown()
 {
     int row = m_view->currentIndex().row();
-    if (row + 1 < m_model.rowCount())
+    if (row + 1 < m_model.rowCount()) {
         MAIN.undoStack()->push(new Playlist::MoveCommand(m_model, row, row + 1));
+        resetPlaylistIndex();
+    }
 }
 
 void PlaylistDock::on_menuButton_clicked()
@@ -277,25 +321,32 @@ void PlaylistDock::on_menuButton_clicked()
     QMenu menu(this);
     QModelIndex index = m_view->currentIndex();
     if (index.isValid() && m_model.playlist()) {
+        menu.addAction(ui->actionOpen);
         if (!MAIN.isMultitrackValid())
             menu.addAction(ui->actionGoto);
+        menu.addAction(ui->actionRemove);
+        menu.addAction(ui->actionCopy);
         if (MLT.isClip())
             menu.addAction(ui->actionInsertCut);
-        menu.addAction(ui->actionOpen);
-        menu.addAction(ui->actionCopy);
         menu.addAction(ui->actionUpdate);
-        menu.addAction(ui->actionRemove);
+        menu.addAction(ui->actionUpdateThumbnails);
+        menu.addAction(ui->actionSetFileDate);
         menu.addSeparator();
     }
     menu.addAction(ui->actionRemoveAll);
+    menu.addAction(ui->actionSelectAll);
+    menu.addAction(ui->actionSelectNone);
+    menu.addAction(ui->actionAddToTimeline);
+    menu.addAction(ui->actionAddToSlideshow);
+    menu.addSeparator();
+
     QMenu* sortByMenu = menu.addMenu(tr("Sort"));
     QActionGroup sortGroup(this);
     sortGroup.addAction(ui->actionSortByName);
     sortGroup.addAction(ui->actionSortByDate);
     sortByMenu->addActions(sortGroup.actions());
-    menu.addAction(ui->actionAddToTimeline);
-    menu.addSeparator();
 
+    menu.addSeparator();
     QMenu* viewModeMenu = menu.addMenu(tr("View mode"));
     QActionGroup modeGroup(this);
     modeGroup.addAction(ui->actionDetailed);
@@ -303,7 +354,6 @@ void PlaylistDock::on_menuButton_clicked()
     modeGroup.addAction(ui->actionIcons);
     viewModeMenu->addActions(modeGroup.actions());
 
-    menu.addSeparator();
     QMenu* subMenu = menu.addMenu(tr("Thumbnails"));
     QActionGroup group(this);
     group.addAction(ui->actionThumbnailsHidden);
@@ -312,6 +362,9 @@ void PlaylistDock::on_menuButton_clicked()
     group.addAction(ui->actionLeftAndRight);
     group.addAction(ui->actionTopAndBottom);
     subMenu->addActions(group.actions());
+
+    menu.addAction(ui->actionPlayAfterOpen);
+
     menu.exec(mapToGlobal(pos));
 }
 
@@ -320,29 +373,32 @@ void PlaylistDock::on_actionInsertCut_triggered()
     if (MLT.isClip() || MLT.savedProducer()) {
         QMimeData mimeData;
         mimeData.setData(Mlt::XmlMimeType, MLT.XML(
-            MLT.isClip()? 0 : MLT.savedProducer()).toUtf8());
+            MLT.isClip()? nullptr : MLT.savedProducer()).toUtf8());
         onDropped(&mimeData, m_view->currentIndex().row());
     }
 }
 
 void PlaylistDock::on_actionAppendCut_triggered()
 {
-    if (MLT.producer() && MLT.producer()->is_valid() && !MAIN.isSourceClipMyProject()) {
+    Mlt::Producer producer(MLT.isClip()? MLT.producer() : MLT.savedProducer());
+    if (producer.is_valid() && !MAIN.isSourceClipMyProject()) {
         if (MLT.isSeekableClip()
             || (MLT.savedProducer() && MLT.isSeekable(MLT.savedProducer()))) {
+            ProxyManager::generateIfNotExists(producer);
             MAIN.undoStack()->push(
-                new Playlist::AppendCommand(m_model,
-                    MLT.XML(MLT.isClip()? 0 : MLT.savedProducer())));
-            MLT.producer()->set(kPlaylistIndexProperty, m_model.playlist()->count());
-            setUpdateButtonEnabled(false);
+                new Playlist::AppendCommand(m_model, MLT.XML(&producer)));
+            setPlaylistIndex(&producer, m_model.playlist()->count() - 1);
+            setUpdateButtonEnabled(true);
         } else {
             DurationDialog dialog(this);
             dialog.setDuration(MLT.profile().fps() * 5);
             if (dialog.exec() == QDialog::Accepted) {
-                MLT.producer()->set_in_and_out(0, dialog.duration() - 1);
-                if (MLT.producer()->get("mlt_service") && !strcmp(MLT.producer()->get("mlt_service"), "avformat"))
-                    MLT.producer()->set("mlt_service", "avformat-novalidate");
+                producer.set_in_and_out(0, dialog.duration() - 1);
+                if (producer.get("mlt_service") && !strcmp(producer.get("mlt_service"), "avformat"))
+                    producer.set("mlt_service", "avformat-novalidate");
                 MAIN.undoStack()->push(new Playlist::AppendCommand(m_model, MLT.XML()));
+                setPlaylistIndex(&producer, m_model.playlist()->count() - 1);
+                setUpdateButtonEnabled(true);
             }
         }
     }
@@ -372,13 +428,16 @@ void PlaylistDock::on_actionAppendBlank_triggered()
 void PlaylistDock::on_actionUpdate_triggered()
 {
     QModelIndex index = m_view->currentIndex();
-    if (!index.isValid() || !m_model.playlist()) return;
-    Mlt::ClipInfo* info = m_model.playlist()->clip_info(index.row());
-    if (!info || MAIN.isSourceClipMyProject()) return;
-    if (MLT.producer()->type() != playlist_type) {
-        if (MLT.isSeekable()) {
-            MAIN.undoStack()->push(new Playlist::UpdateCommand(m_model, MLT.XML(), index.row()));
-            MLT.producer()->set(kPlaylistIndexProperty, index.row() + 1);
+    if (!index.isValid() || !m_model.playlist() || MAIN.isSourceClipMyProject()) return;
+    QScopedPointer<Mlt::ClipInfo> info(m_model.playlist()->clip_info(index.row()));
+    Mlt::Producer producer(MLT.isClip()? MLT.producer() : MLT.savedProducer());
+    if (!info || !producer.is_valid()) return;
+    if (producer.type() != mlt_service_playlist_type) {
+        if (MLT.isSeekableClip()
+                || (MLT.savedProducer() && MLT.isSeekable(MLT.savedProducer()))) {
+            ProxyManager::generateIfNotExists(producer);
+            MAIN.undoStack()->push(new Playlist::UpdateCommand(m_model, MLT.XML(&producer), index.row()));
+            setPlaylistIndex(&producer, index.row());
             setUpdateButtonEnabled(true);
         }
         else {
@@ -386,10 +445,12 @@ void PlaylistDock::on_actionUpdate_triggered()
             DurationDialog dialog(this);
             dialog.setDuration(info->frame_count);
             if (dialog.exec() == QDialog::Accepted) {
-                MLT.producer()->set_in_and_out(0, dialog.duration() - 1);
-                if (MLT.producer()->get("mlt_service") && !strcmp(MLT.producer()->get("mlt_service"), "avformat"))
-                    MLT.producer()->set("mlt_service", "avformat-novalidate");
+                producer.set_in_and_out(0, dialog.duration() - 1);
+                if (producer.get("mlt_service") && !strcmp(producer.get("mlt_service"), "avformat"))
+                    producer.set("mlt_service", "avformat-novalidate");
                 MAIN.undoStack()->push(new Playlist::UpdateCommand(m_model, MLT.XML(), index.row()));
+                setPlaylistIndex(&producer, index.row());
+                setUpdateButtonEnabled(true);
             }
         }
     }
@@ -397,29 +458,36 @@ void PlaylistDock::on_actionUpdate_triggered()
         emit showStatusMessage(tr("You cannot insert a playlist into a playlist!"));
         setUpdateButtonEnabled(false);
     }
-    delete info;
 }
 
 void PlaylistDock::on_removeButton_clicked()
 {
-    QModelIndex index = m_view->currentIndex();
-    if (!index.isValid() || !m_model.playlist()) return;
-    MAIN.undoStack()->push(new Playlist::RemoveCommand(m_model, index.row()));
-    int count = m_model.playlist()->count();
-    if (count == 0) return;
-    int i = index.row() >= count? count-1 : index.row();
-    QScopedPointer<Mlt::ClipInfo> info(m_model.playlist()->clip_info(i));
-    if (info) {
-        emit itemActivated(info->start);
-        int j = MLT.producer()->get_int(kPlaylistIndexProperty);
-        if (j > i + 1) {
-            MLT.producer()->set(kPlaylistIndexProperty, j - 1);
-        } else if (j == i + 1) {
-            // Remove the playlist index property on the producer.
-            MLT.producer()->set(kPlaylistIndexProperty, 0, 0);
-            setUpdateButtonEnabled(false);
+    if (!m_model.playlist() || !m_view->selectionModel()) return;
+    QList<int> rowsRemoved;
+    int n = m_view->selectionModel()->selectedIndexes().size();
+    if (n > 1)
+        MAIN.undoStack()->beginMacro(tr("Remove %n playlist items", nullptr, n));
+    foreach (auto index, m_view->selectionModel()->selectedIndexes()) {
+        int row = index.row();
+        if (!rowsRemoved.contains(row)) {
+            int adjustment = 0;
+            foreach (int i, rowsRemoved) {
+                if (row > i)
+                    --adjustment;
+            }
+            row += adjustment;
+            rowsRemoved << index.row();
+            if (m_model.playlist()->clip_length(row) > 0)
+                MAIN.undoStack()->push(new Playlist::RemoveCommand(m_model, row));
         }
     }
+    if (n > 1)
+        MAIN.undoStack()->endMacro();
+    if (rowsRemoved.contains(MLT.producer()->get_int(kPlaylistIndexProperty))) {
+        // Remove the playlist index property on the producer.
+        resetPlaylistIndex();
+        setUpdateButtonEnabled(false);
+    }
 }
 
 void PlaylistDock::on_actionSetFileDate_triggered()
@@ -432,16 +500,13 @@ void PlaylistDock::on_actionSetFileDate_triggered()
     QScopedPointer<Mlt::ClipInfo> info(m_model.playlist()->clip_info(i));
     if (info && info->producer && info->producer->is_valid()) {
         QString title = info->producer->get("mlt_service");
-        QString resource = QString::fromUtf8(info->producer->get("resource"));
+        QString resource = ProxyManager::resource(*info->producer);
         QFileInfo fileInfo(resource);
-        if (!fileInfo.exists()) {
-            resource = QString::fromUtf8(info->producer->get("warp_resource"));
-            fileInfo = QFileInfo(resource);
-        }
         if (fileInfo.exists()) {
            title = fileInfo.baseName();
         }
         FileDateDialog dialog(resource, info->producer, this);
+        dialog.setWindowModality(QmlApplication::dialogModality());
         dialog.exec();
     }
 }
@@ -455,16 +520,35 @@ void PlaylistDock::onProducerOpened()
 {
     if (!MLT.isMultitrack())
         ui->addButton->setEnabled(true);
+    if (MLT.producer() && MLT.producer()->is_valid()) {
+        auto row = MLT.producer()->get_int(kPlaylistIndexProperty) - 1;
+        if (row < 0 && m_model.rowCount() > 0) {
+            resetPlaylistIndex();
+            emit m_model.dataChanged(m_model.createIndex(0, PlaylistModel::COLUMN_THUMBNAIL),
+                m_model.createIndex(m_model.playlist()->count() - 1, PlaylistModel::COLUMN_THUMBNAIL),
+                QVector<int>() << PlaylistModel::COLUMN_THUMBNAIL);
+        }
+    }
 }
 
 void PlaylistDock::onInChanged()
 {
-    m_inChangedTimer.start(kInOutChangedTimeoutMs);
+    // Order of in/out timers can be important, resolve the other first
+    if (m_outChangedTimer.isActive()) {
+        m_outChangedTimer.stop();
+        onOutTimerFired();
+    }
+    m_inChangedTimer.start();
 }
 
 void PlaylistDock::onOutChanged()
 {
-    m_outChangedTimer.start(kInOutChangedTimeoutMs);
+    // Order of in/out timers can be important, resolve the other first
+    if (m_inChangedTimer.isActive()) {
+        m_inChangedTimer.stop();
+        onInTimerFired();
+    }
+    m_outChangedTimer.start();
 }
 
 void PlaylistDock::on_actionOpen_triggered()
@@ -475,9 +559,10 @@ void PlaylistDock::on_actionOpen_triggered()
     if (i) {
         Mlt::Producer* p = new Mlt::Producer(i->producer);
         p->set_in_and_out(i->frame_in, i->frame_out);
-        p->set(kPlaylistIndexProperty, index.row() + 1);
-        emit clipOpened(p);
+        setPlaylistIndex(p, index.row());
+        emit clipOpened(p, Settings.playlistAutoplay());
         delete i;
+        m_iconsView->resetMultiSelect();
     }
 }
 
@@ -486,22 +571,16 @@ void PlaylistDock::viewCustomContextMenuRequested(const QPoint &pos)
     QModelIndex index = m_view->currentIndex();
     if (index.isValid() && m_model.playlist()) {
         QMenu menu(this);
+        menu.addAction(ui->actionOpen);
         if (!MAIN.isMultitrackValid())
             menu.addAction(ui->actionGoto);
+        menu.addAction(ui->actionRemove);
+        menu.addAction(ui->actionCopy);
         if (MLT.isClip())
             menu.addAction(ui->actionInsertCut);
-        menu.addAction(ui->actionOpen);
-        menu.addAction(ui->actionCopy);
-
-        QScopedPointer<Mlt::ClipInfo> info(m_model.playlist()->clip_info(index.row()));
-        if (info && MLT.producer()->get_int(kPlaylistIndexProperty) == index.row() + 1) {
-            if (MLT.producer()->get_in() != info->frame_in || MLT.producer()->get_out() != info->frame_out)
-                menu.addAction(ui->actionUpdate);
-        }
-
-        menu.addAction(ui->actionRemove);
+        menu.addAction(ui->actionUpdate);
+        menu.addAction(ui->actionUpdateThumbnails);
         menu.addAction(ui->actionSetFileDate);
-
         menu.exec(mapToGlobal(pos));
     }
 }
@@ -517,12 +596,11 @@ void PlaylistDock::viewDoubleClicked(const QModelIndex &index)
         } else {
             Mlt::Producer* p = new Mlt::Producer(i->producer);
             p->set_in_and_out(i->frame_in, i->frame_out);
-            p->set(kPlaylistIndexProperty, index.row() + 1);
-            emit clipOpened(p);
+            setPlaylistIndex(p, index.row());
+            emit clipOpened(p, Settings.playlistAutoplay());
         }
         delete i;
-    } else {
-        MAIN.openVideo();
+        m_iconsView->resetMultiSelect();
     }
 }
 
@@ -533,21 +611,25 @@ void PlaylistDock::on_actionGoto_triggered()
     if (i) {
         emit itemActivated(i->start);
         delete i;
+        m_iconsView->resetMultiSelect();
     }
 }
 
 void PlaylistDock::on_actionRemoveAll_triggered()
 {
+    resetPlaylistIndex();
     MAIN.undoStack()->push(new Playlist::ClearCommand(m_model));
 }
 
 void PlaylistDock::on_actionSortByName_triggered()
 {
+    resetPlaylistIndex();
     MAIN.undoStack()->push(new Playlist::SortCommand(m_model, PlaylistModel::COLUMN_RESOURCE, Qt::AscendingOrder));
 }
 
 void PlaylistDock::on_actionSortByDate_triggered()
 {
+    resetPlaylistIndex();
     MAIN.undoStack()->push(new Playlist::SortCommand(m_model, PlaylistModel::COLUMN_DATE, Qt::AscendingOrder));
 }
 
@@ -555,7 +637,6 @@ void PlaylistDock::onPlaylistCreated()
 {
     ui->removeButton->setEnabled(true);
     ui->updateButton->setEnabled(false);
-    ui->menuButton->setEnabled(true);
     ui->stackedWidget->setCurrentIndex(1);
 }
 
@@ -569,14 +650,15 @@ void PlaylistDock::onPlaylistModified()
 {
     if (m_model.rowCount() == 1)
         ui->tableView->resizeColumnsToContents();
+    if (m_model.rowCount() > 0) {
+        ui->removeButton->setEnabled(true);
+    }
 }
 
 void PlaylistDock::onPlaylistCleared()
 {
     ui->removeButton->setEnabled(false);
     ui->updateButton->setEnabled(false);
-    ui->menuButton->setEnabled(false);
-    ui->stackedWidget->setCurrentIndex(0);
 }
 
 void PlaylistDock::onPlaylistClosed()
@@ -586,12 +668,16 @@ void PlaylistDock::onPlaylistClosed()
 
 void PlaylistDock::onDropped(const QMimeData *data, int row)
 {
+    bool resetIndex = true;
     if (data && data->hasUrls()) {
+        LongUiTask longTask(tr("Add Files"));
         int insertNextAt = row;
         bool first = true;
         QStringList fileNames = Util::sortedFileList(Util::expandDirectories(data->urls()));
-        foreach (QString path, fileNames) {
+        auto i = 0, count = fileNames.size();
+        for (const auto& path : fileNames) {
             if (MAIN.isSourceClipMyProject(path)) continue;
+            longTask.reportProgress(Util::baseName(path), i++, count);
             Mlt::Producer p(MLT.profile(), path.toUtf8().constData());
             if (p.is_valid()) {
                 // Convert MLT XML to a virtual clip.
@@ -601,14 +687,19 @@ void PlaylistDock::onDropped(const QMimeData *data, int row)
                     first = false;
                 }
                 Mlt::Producer* producer = &p;
-                // Convert avformat to avformat-novalidate so that XML loads faster.
-                if (!qstrcmp(producer->get("mlt_service"), "avformat")) {
-                    producer->set("mlt_service", "avformat-novalidate");
-                    producer->set("mute_on_pause", 0);
+                if (first) {
+                    first = false;
+                    if (!MLT.producer() || !MLT.producer()->is_valid()) {
+                        MAIN.open(path, nullptr, false);
+                        if (MLT.producer() && MLT.producer()->is_valid()) {
+                            producer = MLT.producer();
+                            first = true;
+                        }
+                    }
                 }
-                MLT.setImageDurationFromDefault(producer);
-                MLT.lockCreationTime(producer);
+                producer = MLT.setupNewProducer(producer);
                 if (MLT.isSeekable(producer)) {
+                    ProxyManager::generateIfNotExists(*producer);
                     if (row == -1)
                         MAIN.undoStack()->push(new Playlist::AppendCommand(m_model, MLT.XML(producer)));
                     else
@@ -627,24 +718,27 @@ void PlaylistDock::onDropped(const QMimeData *data, int row)
                 if (first) {
                     first = false;
                     setIndex(0);
-                    on_actionOpen_triggered();
+                    resetIndex = false;
                 }
+                delete producer;
             }
         }
     }
     else if (data && data->hasFormat(Mlt::XmlMimeType)) {
         if (MLT.producer() && MLT.producer()->is_valid()) {
-            if (MLT.producer()->type() == playlist_type) {
+            if (MLT.producer()->type() == mlt_service_playlist_type) {
                 emit showStatusMessage(tr("You cannot insert a playlist into a playlist!"));
             } else if (MAIN.isSourceClipMyProject()) {
                 return;
             } else if (MLT.isSeekable()) {
+                Mlt::Producer p(MLT.profile(), "xml-string", data->data(Mlt::XmlMimeType).constData());
+                ProxyManager::generateIfNotExists(p);
                 if (row == -1) {
-                    MAIN.undoStack()->push(new Playlist::AppendCommand(m_model, data->data(Mlt::XmlMimeType)));
-                    MLT.producer()->set(kPlaylistIndexProperty, m_model.playlist()->count());
+                    MAIN.undoStack()->push(new Playlist::AppendCommand(m_model, MLT.XML(&p)));
+                    setPlaylistIndex(MLT.producer(), m_model.playlist()->count() - 1);
                 } else {
-                    MAIN.undoStack()->push(new Playlist::InsertCommand(m_model, data->data(Mlt::XmlMimeType), row));
-                    MLT.producer()->set(kPlaylistIndexProperty, row + 1);
+                    MAIN.undoStack()->push(new Playlist::InsertCommand(m_model, MLT.XML(&p), row));
+                    setPlaylistIndex(MLT.producer(), row);
                 }
                 setUpdateButtonEnabled(true);
             } else {
@@ -662,11 +756,15 @@ void PlaylistDock::onDropped(const QMimeData *data, int row)
             }
         }
     }
+    if (resetIndex)
+        resetPlaylistIndex();
 }
 
 void PlaylistDock::onMoveClip(int from, int to)
 {
     MAIN.undoStack()->push(new Playlist::MoveCommand(m_model, from, to));
+    m_view->clearSelection();
+    resetPlaylistIndex();
 }
 
 void PlaylistDock::onPlayerDragStarted()
@@ -739,7 +837,48 @@ void PlaylistDock::on_actionInOnlyLarge_triggered(bool checked)
 
 void PlaylistDock::on_actionAddToTimeline_triggered()
 {
-    emit addAllTimeline(m_model.playlist());
+    const QModelIndexList& indexes = m_view->selectionModel()->selectedIndexes();
+    Mlt::Playlist playlist(MLT.profile());
+    foreach (auto index, indexes) {
+        if (index.column()) continue;
+        QScopedPointer<Mlt::ClipInfo> info(m_model.playlist()->clip_info(index.row()));
+        if (info && info->producer) {
+            playlist.append(*info->producer, info->frame_in, info->frame_out);
+        }
+    }
+    emit addAllTimeline(&playlist);
+}
+
+void PlaylistDock::on_actionAddToSlideshow_triggered()
+{
+    const QModelIndexList& indexes = m_view->selectionModel()->selectedIndexes();
+    Mlt::Playlist playlist(MLT.profile());
+    foreach (auto index, indexes) {
+        if (index.column()) continue;
+        QScopedPointer<Mlt::ClipInfo> info(m_model.playlist()->clip_info(index.row()));
+        if (info && info->producer) {
+            playlist.append(*info->producer, info->frame_in, info->frame_out);
+        }
+    }
+    if (playlist.count() <= 0 )
+    {
+        return;
+    }
+
+    SlideshowGeneratorDialog dialog(this, playlist);
+    dialog.setWindowModality(QmlApplication::dialogModality());
+    if (dialog.exec() == QDialog::Accepted ) {
+        LongUiTask longTask(QObject::tr("Generate Slideshow"));
+        Mlt::Playlist* slideshow = longTask.runAsync<Mlt::Playlist*>(tr("Generating"), &dialog, &SlideshowGeneratorDialog::getSlideshow);
+        if (slideshow)
+        {
+            if ( slideshow->count() > 0 )
+            {
+                emit addAllTimeline(slideshow, /* skipProxy */ true);
+            }
+            delete slideshow;
+        }
+    }
 }
 
 void PlaylistDock::on_updateButton_clicked()
@@ -751,6 +890,17 @@ void PlaylistDock::on_updateButton_clicked()
     }
 }
 
+void PlaylistDock::onProducerChanged(Mlt::Producer* producer)
+{
+    if (!producer || !producer->is_valid())
+        return;
+    int index = producer->get_int(kPlaylistIndexProperty) - 1;
+    if (index < 0 || !m_model.playlist() || !m_model.playlist()->is_valid() || index >= m_model.playlist()->count())
+        return;
+    MAIN.undoStack()->push(new Playlist::UpdateCommand(m_model, MLT.XML(producer), index));
+    setUpdateButtonEnabled(false);
+}
+
 void PlaylistDock::updateViewModeFromActions()
 {
     PlaylistModel::ViewMode mode;
@@ -779,16 +929,16 @@ void PlaylistDock::setViewMode(PlaylistModel::ViewMode mode)
     if (m_model.viewMode() == mode)
         return;
 
-    ui->listView->setModel(0);
-    ui->tableView->setModel(0);
-    m_iconsView->setModel(0);
+    ui->listView->setModel(nullptr);
+    ui->tableView->setModel(nullptr);
+    m_iconsView->setModel(nullptr);
     ui->listView->hide();
     ui->tableView->hide();
     m_iconsView->hide();
 
     if (ui->listView->itemDelegate()) {
         QAbstractItemDelegate * delegate = ui->listView->itemDelegate();
-        ui->listView->setItemDelegate(0);
+        ui->listView->setItemDelegate(nullptr);
         delete delegate;
     }
 
@@ -814,6 +964,36 @@ void PlaylistDock::setViewMode(PlaylistModel::ViewMode mode)
     m_model.refreshThumbnails();
 }
 
+void PlaylistDock::resetPlaylistIndex()
+{
+    if (MLT.producer())
+        MLT.producer()->set(kPlaylistIndexProperty, nullptr, 0);
+    // Clear the old values
+    for (int j = 0; j < m_model.playlist()->count(); ++j) {
+        Mlt::Producer clip(m_model.playlist()->get_clip(j));
+        clip.parent().Mlt::Properties::clear(kPlaylistIndexProperty);
+    }
+    setUpdateButtonEnabled(false);
+}
+
+void PlaylistDock::emitDataChanged(const QVector<int> &roles)
+{
+    auto row = MLT.producer()->get_int(kPlaylistIndexProperty) - 1;
+    if (row < 0 || row >= m_model.rowCount()) return;
+    auto index = m_model.createIndex(row, PlaylistModel::COLUMN_RESOURCE);
+    emit m_model.dataChanged(index, index, roles);
+}
+
+void PlaylistDock::setPlaylistIndex(Mlt::Producer* producer, int row)
+{
+    // Clear the old values
+    for (int j = 0; j < m_model.playlist()->count(); ++j) {
+        Mlt::Producer clip(m_model.playlist()->get_clip(j));
+        clip.parent().Mlt::Properties::clear(kPlaylistIndexProperty);
+    }
+    producer->set(kPlaylistIndexProperty, row + 1);
+}
+
 #include "playlistdock.moc"
 
 void PlaylistDock::on_tilesButton_clicked()
@@ -876,6 +1056,8 @@ void PlaylistDock::onOutTimerFired()
 void PlaylistDock::keyPressEvent(QKeyEvent* event)
 {
     QDockWidget::keyPressEvent(event);
+    if (event->key() == Qt::Key_Up || event->key() == Qt::Key_Down)
+        event->accept();
     if (!event->isAccepted())
         MAIN.keyPressEvent(event);
 }
@@ -893,11 +1075,72 @@ void PlaylistDock::on_actionCopy_triggered()
     if (!index.isValid() || !m_model.playlist()) return;
     Mlt::ClipInfo* i = m_model.playlist()->clip_info(index.row());
     if (i) {
+        resetPlaylistIndex();
         QString xml = MLT.XML(i->producer);
         Mlt::Producer* p = new Mlt::Producer(MLT.profile(), "xml-string", xml.toUtf8().constData());
         p->set_in_and_out(i->frame_in, i->frame_out);
-        p->set(kPlaylistIndexProperty, index.row() + 1);
+        QGuiApplication::clipboard()->setText(MLT.XML(p));
         emit clipOpened(p);
         delete i;
+        m_iconsView->resetMultiSelect();
+    }
+}
+
+void PlaylistDock::on_actionPlayAfterOpen_triggered(bool checked)
+{
+    Settings.setPlaylistAutoplay(checked);
+}
+
+void PlaylistDock::on_actionSelectAll_triggered()
+{
+    m_view->selectionModel()->clearSelection();
+    for (int i = 0; i < m_model.rowCount(); i++) {
+        m_view->selectionModel()->select(m_model.index(i, 0), QItemSelectionModel::Select | QItemSelectionModel::Rows);
+    }
+}
+
+void PlaylistDock::on_actionSelectNone_triggered()
+{
+    m_view->selectionModel()->clearSelection();
+}
+
+void PlaylistDock::on_actionUpdateThumbnails_triggered()
+{
+    if (!m_model.playlist()) return;
+    m_view->selectionModel()->clearSelection();
+    for (auto i = 0; i < m_model.rowCount(); i++) {
+        m_model.updateThumbnails(i);
+    }
+}
+
+void PlaylistDock::onProducerModified()
+{
+    if (!m_model.playlist()) return;
+    setUpdateButtonEnabled(true);
+
+    // The clip name may have changed.
+    emitDataChanged(QVector<int>() << PlaylistModel::FIELD_RESOURCE);
+}
+
+void PlaylistDock::on_addFilesButton_clicked()
+{
+    QMimeData mimeData;
+    QList<QUrl> urls;
+
+    QString path = Settings.openPath();
+#ifdef Q_OS_MAC
+    path.append("/*");
+#endif
+    LOG_DEBUG() << Util::getFileDialogOptions();
+    QStringList filenames = QFileDialog::getOpenFileNames(this, tr("Open File"), path,
+                            tr("All Files (*);;MLT XML (*.mlt)"), nullptr, Util::getFileDialogOptions());
+
+    if (filenames.length() > 0) {
+        Settings.setOpenPath(QFileInfo(filenames.first()).path());
+        foreach(const QString& s, filenames) {
+            urls << s;
+        }
+        mimeData.setUrls(urls);
+        onDropped(&mimeData, m_view->currentIndex().row() + 1);
     }
 }
