@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2011-2019 Meltytech, LLC
+ * Copyright (c) 2011-2020 Meltytech, LLC
  *
  * GL shader based on BSD licensed code from Peter Bengtsson:
  * http://www.fourcc.org/source/YUV420P-OpenGL-GLSLang.c
@@ -69,11 +69,13 @@ GLWidget::GLWidget(QObject *parent)
     , m_offset(QPoint(0, 0))
     , m_shareContext(0)
     , m_snapToGrid(true)
+    , m_scrubAudio(false)
 {
     LOG_DEBUG() << "begin";
     m_texture[0] = m_texture[1] = m_texture[2] = 0;
     quickWindow()->setPersistentOpenGLContext(true);
     quickWindow()->setPersistentSceneGraph(true);
+    setAttribute(Qt::WA_AcceptTouchEvents);
     quickWindow()->setClearBeforeRendering(false);
     setResizeMode(QQuickWidget::SizeRootObjectToView);
     QDir importPath = QmlUtilities::qmlDir();
@@ -81,6 +83,8 @@ GLWidget::GLWidget(QObject *parent)
     engine()->addImportPath(importPath.path());
     QmlUtilities::setCommonProperties(rootContext());
     rootContext()->setContextProperty("video", this);
+    m_refreshTimer.setInterval(10);
+    m_refreshTimer.setSingleShot(true);
 
     if (Settings.playerGPU())
         m_glslManager = new Filter(profile(), "glsl.manager");
@@ -92,6 +96,8 @@ GLWidget::GLWidget(QObject *parent)
     connect(quickWindow(), SIGNAL(sceneGraphInitialized()), SLOT(initializeGL()), Qt::DirectConnection);
     connect(quickWindow(), SIGNAL(sceneGraphInitialized()), SLOT(setBlankScene()), Qt::QueuedConnection);
     connect(quickWindow(), SIGNAL(beforeRendering()), SLOT(paintGL()), Qt::DirectConnection);
+    connect(&m_refreshTimer, SIGNAL(timeout()), SLOT(onRefreshTimeout()));
+    connect(this, SIGNAL(rectChanged()), SIGNAL(zoomChanged()));
     LOG_DEBUG() << "end";
 }
 
@@ -130,6 +136,11 @@ void GLWidget::initializeGL()
     LOG_INFO() << "OpenGL renderer" << QString::fromUtf8((const char*) glGetString(GL_RENDERER));
     LOG_INFO() << "OpenGL threaded?" << quickWindow()->openglContext()->supportsThreadedOpenGL();
     LOG_INFO() << "OpenGL ES?" << quickWindow()->openglContext()->isOpenGLES();
+    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &m_maxTextureSize);
+    LOG_INFO() << "OpenGL maximum texture size =" << m_maxTextureSize;
+    GLint dims[2];
+    glGetIntegerv(GL_MAX_VIEWPORT_DIMS, &dims[0]);
+    LOG_INFO() << "OpenGL maximum viewport size =" << dims[0] << "x" << dims[1];
 
     if (m_glslManager && quickWindow()->openglContext()->isOpenGLES()) {
         delete m_glslManager;
@@ -187,7 +198,7 @@ void GLWidget::setBlankScene()
 
 void GLWidget::resizeGL(int width, int height)
 {
-    int x, y, w, h;
+    double x, y, w, h;
     double this_aspect = (double) width / height;
     double video_aspect = profile().dar();
 
@@ -209,8 +220,8 @@ void GLWidget::resizeGL(int width, int height)
         w = height * video_aspect;
         h = height;
     }
-    x = (width - w) / 2;
-    y = (height - h) / 2;
+    x = (width - w) / 2.0;
+    y = (height - h) / 2.0;
     m_rect.setRect(x, y, w, h);
     emit rectChanged();
 }
@@ -250,9 +261,9 @@ void GLWidget::createShader()
                                           "varying highp vec2 coordinates;"
                                           "void main(void) {"
                                           "  mediump vec3 texel;"
-                                          "  texel.r = texture2D(Ytex, coordinates).r - 0.0625;" // Y
-                                          "  texel.g = texture2D(Utex, coordinates).r - 0.5;"    // U
-                                          "  texel.b = texture2D(Vtex, coordinates).r - 0.5;"    // V
+                                          "  texel.r = texture2D(Ytex, coordinates).r -  16.0/255.0;" // Y
+                                          "  texel.g = texture2D(Utex, coordinates).r - 128.0/255.0;" // U
+                                          "  texel.b = texture2D(Vtex, coordinates).r - 128.0/255.0;" // V
                                           "  mediump mat3 coefficients;"
                                           "  if (colorspace == 601) {"
                                           "    coefficients = mat3("
@@ -283,7 +294,7 @@ static void uploadTextures(QOpenGLContext* context, SharedFrame& frame, GLuint t
 {
     int width = frame.get_image_width();
     int height = frame.get_image_height();
-    const uint8_t* image = frame.get_image();
+    const uint8_t* image = frame.get_image(mlt_image_yuv420p);
     QOpenGLFunctions* f = context->functions();
 
     // The planes of pixel data may not be a multiple of the default 4 bytes.
@@ -344,8 +355,8 @@ void GLWidget::paintGL()
 #ifndef QT_NO_DEBUG
     QOpenGLFunctions* f = quickWindow()->openglContext()->functions();
 #endif
-    int width = this->width() * devicePixelRatio();
-    int height = this->height() * devicePixelRatio();
+    float width = this->width() * devicePixelRatioF();
+    float height = this->height() * devicePixelRatioF();
 
     glDisable(GL_BLEND);
     glDisable(GL_DEPTH_TEST);
@@ -368,7 +379,7 @@ void GLWidget::paintGL()
     } else if (m_glslManager) {
         m_mutex.lock();
         if (m_sharedFrame.is_valid()) {
-            m_texture[0] = *((GLuint*) m_sharedFrame.get_image());
+            m_texture[0] = *((GLuint*) m_sharedFrame.get_image(mlt_image_opengl_texture));
         }
     }
 
@@ -409,8 +420,8 @@ void GLWidget::paintGL()
     QMatrix4x4 modelView;
     if (m_zoom > 0.0) {
         if (offset().x() || offset().y())
-            modelView.translate(-offset().x() * devicePixelRatio(),
-                                 offset().y() * devicePixelRatio());
+            modelView.translate(-offset().x() * devicePixelRatioF(),
+                                 offset().y() * devicePixelRatioF());
         modelView.scale(zoom(), zoom());
     }
     m_shader->setUniformValue(m_modelViewLocation, modelView);
@@ -418,12 +429,12 @@ void GLWidget::paintGL()
 
     // Provide vertices of triangle strip.
     QVector<QVector2D> vertices;
-    width = m_rect.width() * devicePixelRatio();
-    height = m_rect.height() * devicePixelRatio();
-    vertices << QVector2D(float(-width)/2.0f, float(-height)/2.0f);
-    vertices << QVector2D(float(-width)/2.0f, float( height)/2.0f);
-    vertices << QVector2D(float( width)/2.0f, float(-height)/2.0f);
-    vertices << QVector2D(float( width)/2.0f, float( height)/2.0f);
+    width = m_rect.width() * devicePixelRatioF();
+    height = m_rect.height() * devicePixelRatioF();
+    vertices << QVector2D(-width/2.0f, -height/2.0f);
+    vertices << QVector2D(-width/2.0f, height/2.0f);
+    vertices << QVector2D(width/2.0f, -height/2.0f);
+    vertices << QVector2D(width/2.0f, height/2.0f);
     m_shader->enableAttributeArray(m_vertexLocation);
     check_error(f);
     m_shader->setAttributeArray(m_vertexLocation, vertices.constData());
@@ -464,6 +475,11 @@ void GLWidget::paintGL()
     }
 }
 
+void GLWidget::onRefreshTimeout()
+{
+    Controller::refreshConsumer(m_scrubAudio);
+}
+
 void GLWidget::mousePressEvent(QMouseEvent* event)
 {
     QQuickWidget::mousePressEvent(event);
@@ -478,7 +494,7 @@ void GLWidget::mouseMoveEvent(QMouseEvent* event)
 {
     QQuickWidget::mouseMoveEvent(event);
     if (event->isAccepted()) return;
-    if (event->modifiers() == Qt::ShiftModifier && m_producer) {
+    if (event->modifiers() == (Qt::ShiftModifier | Qt::AltModifier) && m_producer) {
         emit seekTo(m_producer->get_length() * event->x() / width());
         return;
     }
@@ -545,22 +561,28 @@ void GLWidget::createThread(RenderThread **thread, thread_function_t function, v
     (*thread)->start();
 }
 
-static void onThreadCreate(mlt_properties owner, GLWidget* self,
-    RenderThread** thread, int* priority, thread_function_t function, void* data )
+static void onThreadCreate(mlt_properties owner, GLWidget* self, mlt_event_data data)
 {
     Q_UNUSED(owner)
-    Q_UNUSED(priority)
-    self->createThread(thread, function, data);
+    auto threadData = (mlt_event_data_thread*) Mlt::EventData(data).to_object();
+    if (threadData) {
+        auto renderThread = (RenderThread*) threadData->thread;
+        self->createThread(&renderThread, threadData->function, threadData->data);
+    }
 }
 
-static void onThreadJoin(mlt_properties owner, GLWidget* self, RenderThread* thread)
+static void onThreadJoin(mlt_properties owner, GLWidget* self, mlt_event_data data)
 {
     Q_UNUSED(owner)
     Q_UNUSED(self)
-    if (thread) {
-        thread->quit();
-        thread->wait();
-        delete thread;
+    auto threadData = (mlt_event_data_thread*) Mlt::EventData(data).to_object();
+    if (threadData) {
+        auto renderThread = (RenderThread*) threadData->thread;
+        if (renderThread) {
+            renderThread->quit();
+            renderThread->wait();
+            delete renderThread;
+        }
     }
 }
 
@@ -625,7 +647,7 @@ int GLWidget::reconfigure(bool isMulti)
     QString serviceName = property("mlt_service").toString();
     if (!m_consumer || !m_consumer->is_valid()) {
         if (serviceName.isEmpty()) {
-            m_consumer.reset(new Mlt::FilteredConsumer(profile(), "sdl2_audio"));
+            m_consumer.reset(new Mlt::FilteredConsumer(previewProfile(), "sdl2_audio"));
             if (m_consumer->is_valid())
                 serviceName = "sdl2_audio";
             else
@@ -633,9 +655,9 @@ int GLWidget::reconfigure(bool isMulti)
             m_consumer.reset();
         }
         if (isMulti)
-            m_consumer.reset(new Mlt::FilteredConsumer(profile(), "multi"));
+            m_consumer.reset(new Mlt::FilteredConsumer(previewProfile(), "multi"));
         else
-            m_consumer.reset(new Mlt::FilteredConsumer(profile(), serviceName.toLatin1().constData()));
+            m_consumer.reset(new Mlt::FilteredConsumer(previewProfile(), serviceName.toLatin1().constData()));
 
         delete m_threadStartEvent;
         m_threadStartEvent = 0;
@@ -643,9 +665,11 @@ int GLWidget::reconfigure(bool isMulti)
         m_threadStopEvent = 0;
 
         delete m_threadCreateEvent;
-        m_threadCreateEvent = m_consumer->listen("consumer-thread-create", this, (mlt_listener) onThreadCreate);
         delete m_threadJoinEvent;
-        m_threadJoinEvent = m_consumer->listen("consumer-thread-join", this, (mlt_listener) onThreadJoin);
+        if (m_glslManager) {
+            m_threadCreateEvent = m_consumer->listen("consumer-thread-create", this, (mlt_listener) onThreadCreate);
+            m_threadJoinEvent = m_consumer->listen("consumer-thread-join", this, (mlt_listener) onThreadJoin);
+        }
     }
     if (m_consumer->is_valid()) {
         // Connect the producer to the consumer - tell it to "run" later
@@ -666,8 +690,10 @@ int GLWidget::reconfigure(bool isMulti)
             m_consumer->set("0.deinterlace_method", property("deinterlace_method").toString().toLatin1().constData());
             m_consumer->set("0.buffer", qMax(25, qRound(profile().fps())));
             m_consumer->set("0.prefill", qMax(1, qRound(profile().fps() / 25.0)));
+            m_consumer->set("0.drop_max", qRound(profile().fps() / 4.0));
             if (property("keyer").isValid())
                 m_consumer->set("0.keyer", property("keyer").toInt());
+            m_consumer->set("0.video_delay", Settings.playerVideoDelayMs());
         }
         else {
             if (!profile().progressive())
@@ -676,8 +702,10 @@ int GLWidget::reconfigure(bool isMulti)
             m_consumer->set("deinterlace_method", property("deinterlace_method").toString().toLatin1().constData());
             m_consumer->set("buffer", qMax(25, qRound(profile().fps())));
             m_consumer->set("prefill", qMax(1, qRound(profile().fps() / 25.0)));
+            m_consumer->set("drop_max", qRound(profile().fps() / 4.0));
             if (property("keyer").isValid())
                 m_consumer->set("keyer", property("keyer").toInt());
+            m_consumer->set("video_delay", Settings.playerVideoDelayMs());
         }
         if (m_glslManager) {
             if (!m_threadStartEvent)
@@ -699,6 +727,12 @@ int GLWidget::reconfigure(bool isMulti)
     return error;
 }
 
+void GLWidget::refreshConsumer(bool scrubAudio)
+{
+    m_refreshTimer.start();
+    m_scrubAudio = scrubAudio;
+}
+
 QPoint GLWidget::offset() const
 {
     if (m_zoom == 0.0) {
@@ -716,21 +750,30 @@ QImage GLWidget::image() const
     }
     SharedFrame frame = m_frameRenderer->getDisplayFrame();
     if (frame.is_valid()) {
-        int width = frame.get_image_width();
-        int height = frame.get_image_height();
-        QImage result(width, height, QImage::Format_ARGB32);
-        Mlt::Frame displayFrame(frame.clone(false, true));
-        mlt_image_format format = mlt_image_rgb24a;
-        const uchar *image = displayFrame.get_image(format, width, height);
+        const uint8_t* image = frame.get_image(mlt_image_rgba);
         if (image) {
-            QImage temp(width, height, QImage::Format_ARGB32);
-            memcpy(temp.scanLine(0), image, width * height * 4);
-            result = temp.rgbSwapped();
+            int width = frame.get_image_width();
+            int height = frame.get_image_height();
+            QImage temp(image, width, height, QImage::Format_RGBA8888);
+            return temp.copy();
         }
-        return result;
-    } else {
-        return QImage();
     }
+    return QImage();
+}
+
+bool GLWidget::imageIsProxy() const
+{
+    bool isProxy = false;
+    SharedFrame frame = m_frameRenderer->getDisplayFrame();
+    if (frame.is_valid()) {
+        Mlt::Producer* frameProducer = frame.get_original_producer();
+        if (frameProducer && frameProducer->is_valid() && frameProducer->get_int(kIsProxyProperty))
+        {
+            isProxy = true;
+        }
+        delete frameProducer;
+    }
+    return isProxy;
 }
 
 void GLWidget::requestImage() const
@@ -783,7 +826,7 @@ void GLWidget::setOffsetY(int y)
 
 void GLWidget::setCurrentFilter(QmlFilter* filter, QmlMetadata* meta)
 {
-    if (meta && QFile::exists(meta->vuiFilePath().toLocalFile())) {
+    if (meta && meta->type() == QmlMetadata::Filter && QFile::exists(meta->vuiFilePath().toLocalFile())) {
         filter->producer().set(kShotcutVuiMetaProperty, 1);
         rootContext()->setContextProperty("filter", filter);
         setSource(meta->vuiFilePath());
@@ -807,11 +850,10 @@ void GLWidget::updateTexture(GLuint yName, GLuint uName, GLuint vName)
 }
 
 // MLT consumer-frame-show event handler
-void GLWidget::on_frame_show(mlt_consumer, void* self, mlt_frame frame_ptr)
+void GLWidget::on_frame_show(mlt_consumer, GLWidget* widget, mlt_event_data data)
 {
-    Mlt::Frame frame(frame_ptr);
-    if (frame.get_int("rendered")) {
-        GLWidget* widget = static_cast<GLWidget*>(self);
+    auto frame = Mlt::EventData(data).to_frame();
+    if (frame.is_valid() && frame.get_int("rendered")) {
         int timeout = (widget->consumer()->get_int("real_time") > 0)? 0: 1000;
         if (widget->m_frameRenderer && widget->m_frameRenderer->semaphore()->tryAcquire(1, timeout)) {
             QMetaObject::invokeMethod(widget->m_frameRenderer, "showFrame", Qt::QueuedConnection, Q_ARG(Mlt::Frame, frame));
@@ -882,23 +924,17 @@ FrameRenderer::~FrameRenderer()
 
 void FrameRenderer::showFrame(Mlt::Frame frame)
 {
-    int width = 0;
-    int height = 0;
-
     if (!Settings.playerGPU()) {
-        // Convert the image format before creating the SharedFrame.
-        mlt_image_format format = mlt_image_yuv420p;
-        int width = 0;
-        int height = 0;
-        frame.get_image(format, width, height);
         m_displayFrame = SharedFrame(frame);
     }
 
     Q_ASSERT(m_surface->surfaceHandle());
     if (m_context && m_context->isValid()) {
         if (Settings.playerGPU()) {
+            int width = 0;
+            int height = 0;
             frame.set("movit.convert.use_texture", 1);
-            mlt_image_format format = mlt_image_glsl_texture;
+            mlt_image_format format = mlt_image_opengl_texture;
             const GLuint* textureId = (GLuint*) frame.get_image(format, width, height);
 
             m_context->makeCurrent(m_surface);
@@ -968,6 +1004,11 @@ void FrameRenderer::showFrame(Mlt::Frame frame)
     }
     emit frameDisplayed(m_displayFrame);
 
+    if (m_imageRequested) {
+        m_imageRequested = false;
+        emit imageReady();
+    }
+
     m_semaphore.release();
 }
 
