@@ -22,6 +22,8 @@
 
 #include <Logger.h>
 
+#include <QTimer>
+
 static const quintptr NO_PARENT_ID = quintptr(-1);
 
 KeyframesModel::KeyframesModel(QObject* parent)
@@ -69,7 +71,7 @@ QVariant KeyframesModel::data(const QModelIndex& index, int role) const
                     switch (role) {
                     case Qt::DisplayRole:
                     case NameRole: {
-                        QString type = tr("Discrete");
+                        QString type = tr("Hold");
                         switch (const_cast<Mlt::Animation&>(animation).key_get_type(index.row())) {
                         case mlt_keyframe_linear:
                             type = tr("Linear");
@@ -80,7 +82,9 @@ QVariant KeyframesModel::data(const QModelIndex& index, int role) const
                         default:
                             break;
                         }
-                        return QString("%1 - %2").arg(m_filter->timeFromFrames(position)).arg(type);
+                        double value = m_filter->getDouble(name, position);
+                        QString units = m_metadata->keyframes()->parameter(m_metadataIndex[index.internalId()])->units();
+                        return QString("%1 - %2\n%3%4").arg(m_filter->timeFromFrames(position)).arg(type).arg(value).arg(units);
                     }
                     case FrameNumberRole:
                         return position;
@@ -94,11 +98,17 @@ QVariant KeyframesModel::data(const QModelIndex& index, int role) const
                         return result;
                     }
                     case MaximumFrameRole: {
-                        int minimum = animation.previous_key(qMax(0, position - 1)) + 1;
-                        int result = animation.next_key(position + 1) - 1;
-                        result = (result < minimum) ? m_filter->duration() : result;
+                        int result = 0;
+                        if (index.row() >= rowCount(index.parent()) - 1) {
+                            // Last Keyframe
+                            result = m_filter->producer().get_out();
+                        } else {
+                            int minimum = (index.row() > 0 && position > 0) ? (animation.previous_key(position - 1) + 1) : 0;
+                            result = animation.next_key(position + 1) - 1;
+                            result = (result < minimum) ? m_filter->producer().get_out() : result;
+                        }
 //                        LOG_DEBUG() << "keyframeIndex" << index.row() << "maximumFrame" << result;
-                        return result - 1;
+                        return result;
                     }
                     default:
                         break;
@@ -118,9 +128,60 @@ QVariant KeyframesModel::data(const QModelIndex& index, int role) const
         case IsCurveRole:
             return m_metadata->keyframes()->parameter(m_metadataIndex[index.row()])->isCurve();
         case MinimumValueRole:
-            return m_metadata->keyframes()->parameter(m_metadataIndex[index.row()])->minimum();
+        {
+            QmlKeyframesParameter* param = m_metadata->keyframes()->parameter(m_metadataIndex[index.row()]);
+            if (param->rangeType() == QmlKeyframesParameter::MinMax) {
+                return m_metadata->keyframes()->parameter(m_metadataIndex[index.row()])->minimum();
+            } else if (param->rangeType() == QmlKeyframesParameter::ClipLength) {
+                return 0.0;
+            }
+            return 0.0;
+        }
         case MaximumValueRole:
-            return m_metadata->keyframes()->parameter(m_metadataIndex[index.row()])->maximum();
+        {
+            QmlKeyframesParameter* param = m_metadata->keyframes()->parameter(m_metadataIndex[index.row()]);
+            if (param->rangeType() == QmlKeyframesParameter::MinMax) {
+                return m_metadata->keyframes()->parameter(m_metadataIndex[index.row()])->maximum();
+            } else if (param->rangeType() == QmlKeyframesParameter::ClipLength) {
+                int length = m_filter->producer().get_length() - m_filter->in();
+                return (double)length / MLT.profile().fps();
+            }
+            return 0.0;
+        }
+        case LowestValueRole:
+        {
+            QmlKeyframesParameter* param = m_metadata->keyframes()->parameter(m_metadataIndex[index.row()]);
+            Mlt::Animation animation = m_filter->getAnimation(param->property());
+            double min = std::numeric_limits<double>::max();
+            if (animation.is_valid()) {
+                for (int i = 0; i < animation.key_count(); i++) {
+                    int frame = animation.key_get_frame(i);
+                    if (frame >= 0) {
+                        double value = m_filter->getDouble(param->property(), frame);
+                        if (value < min) min = value;
+                    }
+                }
+            }
+            if (min == std::numeric_limits<double>::max()) min = 0;
+            return min;
+        }
+        case HighestValueRole:
+        {
+            QmlKeyframesParameter* param = m_metadata->keyframes()->parameter(m_metadataIndex[index.row()]);
+            Mlt::Animation animation = m_filter->getAnimation(param->property());
+            double max = std::numeric_limits<double>::lowest();
+            if (animation.is_valid()) {
+                for (int i = 0; i < animation.key_count(); i++) {
+                    int frame = animation.key_get_frame(i);
+                    if (frame >= 0) {
+                        double value = m_filter->getDouble(param->property(), frame);
+                        if (value > max) max = value;
+                    }
+                }
+            }
+            if (max == std::numeric_limits<double>::lowest()) max = 0;
+            return max;
+        }
         default:
             break;
         }
@@ -158,6 +219,8 @@ QHash<int, QByteArray> KeyframesModel::roleNames() const
     roles[IsCurveRole]      = "isCurve";
     roles[MinimumValueRole] = "minimum";
     roles[MaximumValueRole] = "maximum";
+    roles[LowestValueRole] = "lowest";
+    roles[HighestValueRole] = "highest";
     roles[FrameNumberRole]  = "frame";
     roles[KeyframeTypeRole] = "interpolation";
     roles[NumericValueRole] = "value";
@@ -195,7 +258,9 @@ bool KeyframesModel::remove(int parameterIndex, int keyframeIndex)
     if (m_filter && parameterIndex < m_propertyNames.count()) {
         QString name = m_propertyNames[parameterIndex];
         Mlt::Animation animation = m_filter->getAnimation(name);
-        if (animation.is_valid()) {
+        // Do not allow the user to delete the last keyframe.
+        // Keyframes should be disabled in the filter panel instead
+        if (animation.is_valid() && animation.key_count() > 1) {
             int frame_num = animation.key_get_frame(keyframeIndex);
             error = animation.remove(frame_num);
             if (!error) {
@@ -218,6 +283,7 @@ bool KeyframesModel::remove(int parameterIndex, int keyframeIndex)
                     modelIndex = index(keyframeIndex, 0, index(parameterIndex));
                     emit dataChanged(modelIndex, modelIndex, QVector<int>() << MinimumFrameRole);
                 }
+                emit dataChanged(index(parameterIndex), index(parameterIndex), QVector<int>() << LowestValueRole << HighestValueRole);
                 emit m_filter->changed();
             }
         }
@@ -296,6 +362,7 @@ bool KeyframesModel::setInterpolation(int parameterIndex, int keyframeIndex, Int
                 emit dataChanged(modelIndex, modelIndex, QVector<int>() << KeyframeTypeRole << NameRole);
                 error = false;
                 emit m_filter->changed();
+                emit m_filter->propertyChanged(name.toUtf8().constData());
             }
         }
     }
@@ -304,31 +371,56 @@ bool KeyframesModel::setInterpolation(int parameterIndex, int keyframeIndex, Int
     return error;
 }
 
-bool KeyframesModel::setPosition(int parameterIndex, int keyframeIndex, int position)
+void KeyframesModel::setKeyframePosition(int parameterIndex, int keyframeIndex, int position)
 {
-    bool error = true;
-    if (m_filter && parameterIndex < m_propertyNames.count()) {
-        QString name = m_propertyNames[parameterIndex];
-        Mlt::Animation animation = m_filter->getAnimation(name);
-        if (animation.is_valid()) {
-            if (!animation.key_set_frame(keyframeIndex, position)) {
-//                LOG_DEBUG() << "keyframe index" << keyframeIndex << "position" << position;
-                foreach (name, m_metadata->keyframes()->parameter(m_metadataIndex[parameterIndex])->gangedProperties()) {
-                    Mlt::Animation animation = m_filter->getAnimation(name);
-                    if (animation.is_valid())
-                        animation.key_set_frame(keyframeIndex, position);
-                }
-                QModelIndex modelIndex = index(keyframeIndex, 0, index(parameterIndex));
-                emit dataChanged(modelIndex, modelIndex, QVector<int>() << FrameNumberRole << NameRole);
-                updateNeighborsMinMax(parameterIndex, keyframeIndex);
-                error = false;
-                emit m_filter->changed();
-            }
-        }
+    if (!m_filter) {
+        LOG_ERROR() << "Invalid Filter" << parameterIndex;
+        return;
     }
-    if (error)
-        LOG_ERROR() << "failed to set keyframe" << "at parameter index" << parameterIndex << "keyframeIndex" << keyframeIndex << "to position" << position;
-    return error;
+
+    if (parameterIndex >= m_propertyNames.count()) {
+        LOG_ERROR() << "Invalid parameter index" << parameterIndex;
+        return;
+    }
+
+    QString name = m_propertyNames[parameterIndex];
+    Mlt::Animation animation = m_filter->getAnimation(name);
+    if (!animation.is_valid()) {
+        LOG_ERROR() << "Invalid animation" << parameterIndex;
+        return;
+    }
+
+    if (keyframeIndex >= animation.key_count()) {
+        LOG_ERROR() << "Invalid key index" << parameterIndex << keyframeIndex;
+        return;
+    }
+
+    if (position < 0) {
+        LOG_ERROR() << "Invalid key position" << parameterIndex << keyframeIndex << position;
+        return;
+    }
+
+    int prevPosition = animation.key_get_frame(keyframeIndex);
+    if (position == prevPosition) {
+        LOG_ERROR() << "Position did not change" << parameterIndex << keyframeIndex << position;
+        return;
+    }
+
+    if (animation.key_set_frame(keyframeIndex, position)) {
+        LOG_ERROR() << "Failed to set position" << parameterIndex << keyframeIndex << position;
+        return;
+    }
+
+    foreach (name, m_metadata->keyframes()->parameter(m_metadataIndex[parameterIndex])->gangedProperties()) {
+        Mlt::Animation animation = m_filter->getAnimation(name);
+        if (animation.is_valid())
+            animation.key_set_frame(keyframeIndex, position);
+    }
+    QModelIndex modelIndex = index(keyframeIndex, 0, index(parameterIndex));
+    emit dataChanged(modelIndex, modelIndex, QVector<int>() << FrameNumberRole << NameRole);
+    updateNeighborsMinMax(parameterIndex, keyframeIndex);
+    emit m_filter->changed();
+    emit m_filter->propertyChanged(name.toUtf8().constData());
 }
 
 void KeyframesModel::addKeyframe(int parameterIndex, double value, int position, KeyframesModel::InterpolationType type)
@@ -345,40 +437,140 @@ void KeyframesModel::addKeyframe(int parameterIndex, int position)
 {
     if (m_filter && parameterIndex < m_propertyNames.count()) {
         QString name = m_propertyNames[parameterIndex];
-        // Get the value from the existing position.
-        double value = m_filter->getDouble(name, position);
-        Mlt::Animation anim = m_filter->getAnimation(name);
-        if (anim.is_valid() && !anim.is_key(position)) {
-            mlt_keyframe_type keyframeType = m_filter->getKeyframeType(anim, position, mlt_keyframe_type(-1));
-            // Simply adding a keyframe does not change the current value of
-            // the filter parameter. So, no need to trigger a bunch of signals
-            // and refresh the consumer. Besides, refreshing the consumer is
-            // some how causing the player to seek ahead by one frame inadvertently
-            // such that changing the parameter value causes the addition of a
-            // keyframe just after this one. MLT.refreshConsumer() with
-            // frame-dropping enabled may have dropped video of the most recent
-            // frame from the producer, but Shotcut does not know about it
-            // because it did not receive a "consumer-frame-show" event for it.
-            m_filter->blockSignals(true);
-            m_filter->set(name, value, position, keyframeType);
-            m_filter->blockSignals(false);
-            onFilterChanged(name);
+        auto parameter = m_metadata->keyframes()->parameter(m_metadataIndex[parameterIndex]);
+        if (parameter->isRectangle()) {
+            auto value = m_filter->getRect(name, position);
+            Mlt::Animation anim = m_filter->getAnimation(name);
+            if (anim.is_valid() && !anim.is_key(position)) {
+                mlt_keyframe_type keyframeType = m_filter->getKeyframeType(anim, position, mlt_keyframe_type(-1));
+                m_filter->blockSignals(true);
+                m_filter->set(name, value, 1.0, position, keyframeType);
+                m_filter->blockSignals(false);
+            }
+        } else if (parameter->isCurve()) {
+            // Get the value from the existing position.
+            double value = m_filter->getDouble(name, position);
+            Mlt::Animation anim = m_filter->getAnimation(name);
+            if (anim.is_valid() && !anim.is_key(position)) {
+                mlt_keyframe_type keyframeType = m_filter->getKeyframeType(anim, position, mlt_keyframe_type(-1));
+                // Simply adding a keyframe does not change the current value of
+                // the filter parameter. So, no need to trigger a bunch of signals
+                // and refresh the consumer. Besides, refreshing the consumer is
+                // some how causing the player to seek ahead by one frame inadvertently
+                // such that changing the parameter value causes the addition of a
+                // keyframe just after this one. MLT.refreshConsumer() with
+                // frame-dropping enabled may have dropped video of the most recent
+                // frame from the producer, but Shotcut does not know about it
+                // because it did not receive a "consumer-frame-show" event for it.
+                m_filter->blockSignals(true);
+                m_filter->set(name, value, position, keyframeType);
+                for (auto& key : parameter->gangedProperties())
+                    m_filter->set(key, value, position, keyframeType);
+                m_filter->blockSignals(false);
+            }
+        } else {
+            emit keyframeAdded(name, position);
         }
+        onFilterChanged(name);
     }
 }
 
-void KeyframesModel::setKeyframe(int parameterIndex, double value, int position, KeyframesModel::InterpolationType type)
+void KeyframesModel::setKeyframeValue(int parameterIndex, int keyframeIndex, double value)
 {
-    if (m_filter && parameterIndex < m_propertyNames.count()) {
-        QString name = m_propertyNames[parameterIndex];
-        m_filter->filter().anim_set(name.toUtf8().constData(), value, position, m_filter->duration(), mlt_keyframe_type(type));
-        foreach (name, m_metadata->keyframes()->parameter(m_metadataIndex[parameterIndex])->gangedProperties())
-            m_filter->filter().anim_set(name.toUtf8().constData(), value, position, m_filter->duration(), mlt_keyframe_type(type));
-        emit m_filter->changed();
-        QModelIndex modelIndex = index(keyframeIndex(parameterIndex, position), 0, index(parameterIndex));
-        emit dataChanged(modelIndex, modelIndex, QVector<int>() << NumericValueRole << NameRole);
-        updateNeighborsMinMax(parameterIndex, modelIndex.row());
+    if (!m_filter) {
+        LOG_ERROR() << "Invalid Filter" << parameterIndex;
+        return;
     }
+
+    if (parameterIndex >= m_propertyNames.count()) {
+        LOG_ERROR() << "Invalid parameter index" << parameterIndex;
+        return;
+    }
+
+    QString name = m_propertyNames[parameterIndex];
+    Mlt::Animation animation = m_filter->getAnimation(name);
+    if (!animation.is_valid()) {
+        LOG_ERROR() << "Invalid animation" << parameterIndex;
+        return;
+    }
+
+    if (keyframeIndex >= animation.key_count()) {
+        LOG_ERROR() << "Invalid key index" << parameterIndex << keyframeIndex;
+        return;
+    }
+
+    int position = animation.key_get_frame(keyframeIndex);
+    if (position < 0) {
+        LOG_ERROR() << "Invalid position" << parameterIndex << keyframeIndex;
+        return;
+    }
+
+    mlt_keyframe_type type = animation.key_get_type(keyframeIndex);
+    m_filter->service().anim_set(name.toUtf8().constData(), value, position, m_filter->duration(), type);
+    foreach (name, m_metadata->keyframes()->parameter(m_metadataIndex[parameterIndex])->gangedProperties())
+        m_filter->service().anim_set(name.toUtf8().constData(), value, position, m_filter->duration(), type);
+    emit m_filter->changed();
+    emit m_filter->propertyChanged(name.toUtf8().constData());
+    QModelIndex modelIndex = index(keyframeIndex, 0, index(parameterIndex));
+    emit dataChanged(modelIndex, modelIndex, QVector<int>() << NumericValueRole << NameRole);
+    emit dataChanged(index(parameterIndex), index(parameterIndex), QVector<int>() << LowestValueRole << HighestValueRole);
+}
+
+void KeyframesModel::setKeyframeValuePosition(int parameterIndex, int keyframeIndex, double value, int position)
+{
+    if (!m_filter) {
+        LOG_ERROR() << "Invalid Filter" << parameterIndex;
+        return;
+    }
+
+    if (parameterIndex >= m_propertyNames.count()) {
+        LOG_ERROR() << "Invalid parameter index" << parameterIndex;
+        return;
+    }
+
+    QString name = m_propertyNames[parameterIndex];
+    Mlt::Animation animation = m_filter->getAnimation(name);
+    if (!animation.is_valid()) {
+        LOG_ERROR() << "Invalid animation" << parameterIndex;
+        return;
+    }
+
+    if (keyframeIndex >= animation.key_count()) {
+        LOG_ERROR() << "Invalid key index" << parameterIndex << keyframeIndex;
+        return;
+    }
+
+    if (position < 0) {
+        LOG_ERROR() << "Invalid key position" << parameterIndex << keyframeIndex << position;
+        return;
+    }
+
+    QVector<int> roles;
+    int prevPosition = animation.key_get_frame(keyframeIndex);
+    if (position != prevPosition) {
+        if (animation.key_set_frame(keyframeIndex, position)) {
+            LOG_ERROR() << "Failed to set position" << parameterIndex << keyframeIndex << position;
+            return;
+        }
+        foreach (name, m_metadata->keyframes()->parameter(m_metadataIndex[parameterIndex])->gangedProperties()) {
+            Mlt::Animation animation = m_filter->getAnimation(name);
+            if (animation.is_valid())
+                animation.key_set_frame(keyframeIndex, position);
+        }
+        roles << FrameNumberRole;
+        updateNeighborsMinMax(parameterIndex, keyframeIndex);
+    }
+
+    mlt_keyframe_type type = animation.key_get_type(keyframeIndex);
+    m_filter->service().anim_set(name.toUtf8().constData(), value, position, m_filter->duration(), type);
+    foreach (name, m_metadata->keyframes()->parameter(m_metadataIndex[parameterIndex])->gangedProperties())
+        m_filter->service().anim_set(name.toUtf8().constData(), value, position, m_filter->duration(), type);
+    emit m_filter->changed();
+    emit m_filter->propertyChanged(name.toUtf8().constData());
+    roles << NumericValueRole << NameRole;
+    QModelIndex modelIndex = index(keyframeIndex, 0, index(parameterIndex));
+    emit dataChanged(modelIndex, modelIndex, roles);
+    emit dataChanged(index(parameterIndex), index(parameterIndex), QVector<int>() << LowestValueRole << HighestValueRole);
 }
 
 bool KeyframesModel::isKeyframe(int parameterIndex, int position)
@@ -412,78 +604,76 @@ void KeyframesModel::reload()
 
 void KeyframesModel::onFilterChanged(const QString& property)
 {
-//    LOG_DEBUG() << property;
+    bool isKeyframeProperty = false;
+    for (int p = 0; p < m_metadata->keyframes()->parameterCount() && isKeyframeProperty == false; p++) {
+        if (m_metadata->keyframes()->parameter(p)->property() == property) {
+            isKeyframeProperty = true;
+            break;
+        }
+        QString name;
+        foreach (name, m_metadata->keyframes()->parameter(p)->gangedProperties()) {
+            if (name == property) {
+                isKeyframeProperty = true;
+                break;
+            }
+        }
+    }
+    if (!isKeyframeProperty) {
+        // Does not affect this model.
+        return;
+    }
+
     int i = m_propertyNames.indexOf(property);
-    if (i > -1) {
-        int count = m_keyframeCounts[i];
-        m_keyframeCounts[i] = keyframeCount(i);
-        if (count > 0) {
-            beginRemoveRows(index(i), 0, count - 1);
+    if (i < 0) {
+        // First keyframe added. Reset model to add this parameter.
+        reload();
+        return;
+    }
+
+    int prevCount = m_keyframeCounts[i];
+    m_keyframeCounts[i] = keyframeCount(i);
+    if (m_keyframeCounts[i] == 0) {
+        // All keyframes removed. Reset model to remove this parameter.
+        reload();
+    }
+    else if (prevCount != m_keyframeCounts[i]) {
+        // Keyframe count changed. Remove all old and insert all new.
+        if (prevCount > 0) {
+            beginRemoveRows(index(i), 0, prevCount - 1);
             endRemoveRows();
         }
-        if (m_keyframeCounts[i] > 0) {
-//            LOG_DEBUG() << property << m_filter->get(property) << m_keyframeCounts[i];
-            beginInsertRows(index(i), 0, m_keyframeCounts[i] - 1);
-            endInsertRows();
-        } else {
-            reload();
-        }
-    } else {
-        reload();
+        beginInsertRows(index(i), 0, m_keyframeCounts[i] - 1);
+        endInsertRows();
+        emit dataChanged(index(i), index(i), QVector<int>() << LowestValueRole << HighestValueRole);
+    }
+    else {
+        // Keyframe count is unchanged. A value must have changed.
+        emit dataChanged(index(i), index(i), QVector<int>() << LowestValueRole << HighestValueRole);
+        emit dataChanged(index(0, 0, index(i)), index(m_keyframeCounts[i] - 1, 0, index(i)), QVector<int>() << NumericValueRole);
     }
 }
 
 void KeyframesModel::onFilterInChanged(int delta)
 {
-    for (int parameterIndex = 0; parameterIndex < m_propertyNames.count(); parameterIndex++) {
-        int count = m_keyframeCounts[parameterIndex];
-        if (count > 0) {
-            Mlt::Animation animation = m_filter->getAnimation(m_propertyNames[parameterIndex]);
-            if (animation.is_valid()) {
-                for (int keyframeIndex = 0; keyframeIndex < count;) {
-                    int newFrame = animation.key_get_frame(keyframeIndex) - delta;
-                    if (animation.is_key(newFrame)) {
-                        beginRemoveRows(index(parameterIndex), keyframeIndex, keyframeIndex);
-                        animation.remove(animation.key_get_frame(keyframeIndex));
-                        animation.interpolate();
-                        m_keyframeCounts[parameterIndex] -= 1;
-                        endRemoveRows();
-                        --count;
-                    } else {
-                        animation.key_set_frame(keyframeIndex, newFrame);
-                        ++keyframeIndex;
-                    }
-                }
-                QModelIndex parentIndex = index(parameterIndex);
-                emit dataChanged(index(0, 0, parentIndex), index(count - 1, 0, parentIndex), QVector<int>() << FrameNumberRole);
-            }
-        }
+    QTimer::singleShot(0, this, SLOT(reload()));
+}
+
+
+void KeyframesModel::trimFilterIn(int in)
+{
+    Mlt::Service& service = m_filter->service();
+    if (service.is_valid() && service.type() == mlt_service_filter_type) {
+        Mlt::Filter filter = service;
+        MLT.adjustFilter(&filter, filter.get_in(), filter.get_out(), in - filter.get_in(), 0);
     }
 }
 
-void KeyframesModel::onFilterOutChanged(int delta)
+void KeyframesModel::trimFilterOut(int out)
 {
-    Q_UNUSED(delta)
-    for (int parameterIndex = 0; parameterIndex < m_propertyNames.count(); parameterIndex++) {
-        int count = m_keyframeCounts[parameterIndex];
-        if (count > 0) {
-            Mlt::Animation animation = m_filter->getAnimation(m_propertyNames[parameterIndex]);
-            if (animation.is_valid()) {
-                for (int keyframeIndex = 0; keyframeIndex < count;) {
-                    int frame = animation.key_get_frame(keyframeIndex);
-                    if (frame < 0) {
-                        beginRemoveRows(index(parameterIndex), keyframeIndex, keyframeIndex);
-                        animation.remove(animation.key_get_frame(keyframeIndex));
-                        animation.interpolate();
-                        m_keyframeCounts[parameterIndex] -= 1;
-                        endRemoveRows();
-                        --count;
-                    } else {
-                        ++keyframeIndex;
-                    }
-                }
-            }
-        }
+    Mlt::Service& service = m_filter->service();
+    if (service.is_valid() && service.type() == mlt_service_filter_type) {
+        Mlt::Filter filter = service;
+        MLT.adjustFilter(&filter, filter.get_in(), filter.get_out(), 0, filter.get_out() - out);
     }
 }
 
